"""
Gather stock market data from 3rd party sources
"""

import yfinance as yf
import pandas as pd
from dotenv import load_dotenv
from loguru import logger
import os
from pathlib import Path
import fmpsdk
from fmpsdk.settings import DEFAULT_LIMIT, SEC_RSS_FEEDS_FILENAME, BASE_URL_v3
from fmpsdk.url_methods import __return_json_v4
import typing
import os
from pathlib import Path
from dotenv import load_dotenv
from canswim.hfhub import HFHub
import typing
from fmpsdk.url_methods import __return_json_v3, __validate_period
from pandas.api.types import is_datetime64_any_dtype as is_datetime


def get_latest_date(datetime_col):
    """Return the latest datetime value in a datetime formatted dataframe column"""
    assert len(datetime_col) > 0
    assert is_datetime(datetime_col)
    latest_saved_record = datetime_col.max()
    return latest_saved_record


FMP_API_DEFAULT_LIMIT = -1


def institutional_symbol_ownership(
    apikey: str,
    symbol: str,
    limit: int = FMP_API_DEFAULT_LIMIT,
    includeCurrentQuarter: bool = False,
) -> typing.Optional[typing.List[typing.Dict]]:
    """
    Query FMP /institutional-ownership/ API.

    :param apikey: Your API key.
    :param symbol: Company ticker.
    :param limit: up to how many quarterly reports to return.
    :param includeCurrentQuarter: Whether to include any available data in the current quarter.
    :return: A list of dictionaries.
    """
    path = f"institutional-ownership/symbol-ownership"
    query_vars = {
        "symbol": symbol,
        "apikey": apikey,
        "includeCurrentQuarter": includeCurrentQuarter,
        "limit": limit,
    }
    return __return_json_v4(path=path, query_vars=query_vars)


def analyst_estimates(
    apikey: str, symbol: str, period: str = "annual", limit: int = FMP_API_DEFAULT_LIMIT
) -> typing.Optional[typing.List[typing.Dict]]:
    """
    Query FMP /analyst-estimates/ API.

    :param apikey: Your API key.
    :param symbol: Company ticker.
    :param period: 'annual' or 'quarter'
    :param limit: Number of rows to return.
    :return: A list of dictionaries.
    """
    path = f"/analyst-estimates/{symbol}"
    query_vars = {
        "apikey": apikey,
        "symbol": symbol,
        "period": __validate_period(value=period),
        "limit": limit,
    }
    return __return_json_v3(path=path, query_vars=query_vars)


class MarketDataGatherer:

    def __init__(self) -> None:
        load_dotenv(override=True)
        self.FMP_API_KEY = os.getenv("FMP_API_KEY")
        logger.info(f"FMP_API_KEY found: {self.FMP_API_KEY!= None}")
        self.all_stocks_file = "all_stocks.csv"
        self.price_frequency = "1d"  # "1wk"
        self.min_start_date = os.getenv("train_date_start", "1991-01-01")
        self.data_dir = os.getenv("data_dir", "data")
        self.data_3rd_party = os.getenv("data-3rd-party", "data-3rd-party")

    def gather_stock_tickers(self):
        # Prepare list of stocks for training
        all_stock_set = set()
        stock_files = [
            "IBD50.csv",
            "IBD250.csv",
            "ibdlive_picks.csv",
            "russell2000_iwm_holdings.csv",
            "sp500_ivv_holdings.csv",
            "nasdaq100_cndx_holdings.csv",
            self.all_stocks_file,
        ]
        logger.info(
            "Compiling list of stocks to train on from: {files}", files=stock_files
        )
        for f in stock_files:
            fp = f"data/data-3rd-party/{f}"
            if Path(fp).is_file():
                stocks = pd.read_csv(fp)
                logger.info(f"loaded {len(stocks)} symbols from {fp}")
                stock_set = set(stocks["Symbol"])
                logger.info(f"{len(stock_set)} symbols in stock set")
                all_stock_set |= stock_set
                logger.info(f"total symbols loaded: {len(all_stock_set)}")
            else:
                logger.info(f"{fp} not found.")

        logger.info(f"Total size of stock set: {len(all_stock_set)}")
        # logger.info(f"All stocks set: {all_stock_set}")
        # Drop invalid symbols with more than 10 ticker characters.
        # Almost all stocks have less than 5 characters.
        slist = [x for x in set(all_stock_set) if len(x) < 10]
        self.stocks_ticker_set = set(slist)
        # drop known junk symbols from the data feed
        junk_tickers = set(
            [
                "MSFUT",
                "GEFB",
                "METCV",
                "SGAFT",
                "NQH4",
                "XTSLA",
                "-",
                "PDLI",
                "ADRO",
                "ICSUAGD",
                "BFB",
                "GTXI",
                "P5N994",
                "LGFB",
                "MLIFT",
                "ESH4",
                "LGFA",
                "MOGA",
                "PBRA",
                "BRKB",
                "RTYH4",
                "\xa0",
                "CRDA",
                "GBP",
                "ESMT",
                "MRTX",
                "RAIN",
                "CAD",
                "EUR",
            ]
        )
        self.stocks_ticker_set = self.stocks_ticker_set - junk_tickers
        stocks_df = pd.DataFrame()
        stocks_df["Symbol"] = list(self.stocks_ticker_set)
        stocks_df = stocks_df.set_index(["Symbol"])
        stocks_df.index = stocks_df.index.drop_duplicates()
        stocks_df
        stocks_file = f"data/data-3rd-party/{self.all_stocks_file}"
        stocks_df.to_csv(stocks_file)
        logger.info(f"Saved stock set to {stocks_file}")

    def _gather_yfdata_date_index(self, data_file: str = None, tickers: str = None):
        start_date = self.min_start_date
        old_df = None
        try:
            old_df = pd.read_parquet(data_file)
            logger.info("Loaded saved data. Sample: \n{df}", df=old_df)
            start_date = get_latest_date(old_df.index) - pd.Timedelta(1, "d")
            logger.info(f"Columns: \n{old_df.columns}")
            logger.info(f"Latest saved record is after {start_date}")
        except Exception as e:
            logger.exception(f"Could not load data from file: {data_file}. Error: {e}")
        new_df = yf.download(tickers, start=start_date, group_by="tickers", period="1d")
        new_df = new_df.dropna(how="all")
        logger.info("New data gathered. Sample: \n{bm}", bm=new_df)
        logger.info(f"Columns: \n{new_df.columns}")
        if old_df is not None:
            assert sorted(old_df.columns) == sorted(new_df.columns)
            merged_df = pd.concat([old_df, new_df], axis=0)
            # logger.info(f"bm_df concat\n {merged_df}")
            assert sorted(merged_df.columns) == sorted(old_df.columns)
            assert len(merged_df) == len(old_df) + len(new_df)
            merged_df = merged_df[~merged_df.index.duplicated(keep="last")]
        else:
            merged_df = new_df
        logger.info("Updated data ready. Sample: \n{bm}", bm=merged_df)
        assert merged_df.index.is_unique
        merged_df.to_parquet(data_file)
        logger.info(f"Saved data to {data_file}")
        _bm = pd.read_parquet(data_file)
        assert sorted(_bm.columns) == sorted(merged_df.columns)
        assert len(_bm) == len(merged_df)
        logger.info(f"Sanity check passed for saved data. Loaded OK from {data_file}")

    def gather_broad_market_data(self):
        ## Prepare data for broad market indicies
        # Capture S&P500, NASDAQ100 and Russell 200 indecies and their equal weighted counter parts
        # As well as VIX volatility index, DYX US Dollar index, TNX US 12 Weeks Treasury Yield, 5 Years Treasury Yield and 10 Year Treasuries Yield
        broad_market_indicies = (
            "^SPX ^SPXEW ^NDX ^NDXE ^RUT ^R2ESC ^VIX DX-Y.NYB ^IRX ^FVX ^TNX"
        )
        data_file = "data/data-3rd-party/broad_market.parquet"
        self._gather_yfdata_date_index(
            data_file=data_file, tickers=broad_market_indicies
        )

    def gather_sectors_data(self):
        """Gather historic price and volume data for key market sectors"""
        sector_indicies = "XLE ^SP500-15 ^SP500-20 ^SP500-25 ^SP500-30 ^SP500-35 ^SP500-40 ^SP500-45 ^SP500-50 ^SP500-55 ^SP500-60"
        data_file = "data/data-3rd-party/sectors.parquet"
        self._gather_yfdata_date_index(data_file=data_file, tickers=sector_indicies)

    def gather_subindustries_data(self):
        """
        WARN: YFinance does not provide rich historical price and volume data for Industries and Sub-industries
        the way it does for Sectors.
        Do not use this method with YFinance.
        """
        return  # See warning message above.
        """
        Gather historic price and volume data for S&P 1500 GICS subindustries indecies.
        S&P 1500 includes S&P 400, S&P 500, S&P 600 and overall about 90% of the US stock market capitalization.
        The dataset has 163 GICS sub-industry indecies active as of 2023 plus 7 that were removed in 2023.
        https://www.msci.com/documents/1296102/11185224/GICS+Map+2023.xlsx/82cc6504-9919-29e5-9789-a24fc039d0a5?t=1679087572540
        The goal of these covariates is to provide the model with a more granural breakdown of stock grouping by industry.
        Since stocks usually move together with their group, the model can learn the patterns how an individual stock trend
        relates to its group price and volume action.
        """
        subindustry_indicies = []
        gics_file = f"{self.data_dir}/{self.data_3rd_party}/GICS2023.csv"
        gics_df = pd.read_csv(gics_file)
        logger.info(f"Loaded {len(gics_df)} GICS records. Columns: {gics_df}")
        code_col_name = "Sub-Industry Code"
        subindustry_codes = gics_df[
            pd.to_numeric(gics_df[code_col_name], errors="coerce").notnull()
        ]
        subindustry_codes = subindustry_codes[code_col_name].unique()
        logger.info(
            f"Loaded {len(subindustry_codes)} subindustry codes: {subindustry_codes}"
        )
        sub_prefix = "^sp1500-"
        subindustry_symbols = []
        for c in subindustry_codes:
            subindustry_symbols.append(f"{sub_prefix}{c}")
        logger.info(
            f"Prepared list of {len(subindustry_symbols)} S&P1500 subindustry symbols: {subindustry_symbols}"
        )
        data_file = "data/data-3rd-party/subindustries.parquet"
        self._gather_yfdata_date_index(data_file=data_file, tickers=subindustry_symbols)
        logger.info("Finished gathering subindisty data")

    def gather_stock_price_data(self):
        data_file = (
            f"data/data-3rd-party/all_stocks_price_hist_{self.price_frequency}.parquet"
        )
        start_date = self.min_start_date
        old_df = None
        try:
            old_df = pd.read_parquet(data_file)
            logger.info("Loaded saved data. Sample: \n{df}", df=old_df)
            start_date = get_latest_date(
                old_df.index.get_level_values("Date")
            ) - pd.Timedelta(1, "d")
            logger.info(f"Columns: \n{old_df.columns}")
            logger.info(f"Latest saved record is after {start_date}")
        except Exception as e:
            logger.exception(f"Could not load data from file: {data_file}. Error: {e}")
        new_df = yf.download(
            self.stocks_ticker_set, start=start_date, group_by="tickers"
        )
        new_df = new_df.dropna(how="all")
        new_df = new_df.stack(level=0)
        new_df.index.names = ["Date", "Symbol"]
        new_df.index = new_df.index.swaplevel(0)
        logger.info(f"Stock price data index: {new_df.index.names}")
        logger.info("New data gathered. Sample: \n{df}", df=new_df)
        logger.info(f"Columns: \n{new_df.columns}")
        if old_df is not None:
            assert sorted(old_df.columns) == sorted(new_df.columns)
            merged_df = pd.concat([old_df, new_df], axis=0, join="inner")
            # logger.info(f"merged_df concat\n {merged_df}")
            assert sorted(merged_df.columns) == sorted(old_df.columns)
            assert len(merged_df) == len(old_df) + len(new_df)
            merged_df = merged_df[~merged_df.index.duplicated(keep="last")]
        else:
            merged_df = new_df
        logger.info("Updated data ready. Sample: \n{df}", df=merged_df)
        # df=merged_df.loc[merged_df.index.get_level_values("Symbol") == 'TEAM'])
        assert merged_df.index.is_unique
        merged_df.to_parquet(data_file)
        logger.info(f"Saved broad market data to {data_file}")
        _bm = pd.read_parquet(data_file)
        assert sorted(_bm.columns) == sorted(merged_df.columns)
        assert len(_bm) == len(merged_df)
        logger.info(
            f"Sanity check passed for broad market data. Loaded OK from {data_file}"
        )

    def gather_earnings_data(self):
        logger.info("Gathering  earnings and sales data...")
        earnings_all_df = pd.DataFrame()
        for ticker in self.stocks_ticker_set:  # ['AAON']: #
            earnings = fmpsdk.historical_earning_calendar(
                apikey=self.FMP_API_KEY, symbol=ticker, limit=-1
            )
            if earnings is not None and len(earnings) > 0:
                edf = pd.DataFrame(earnings)
                edf["date"] = pd.to_datetime(edf["date"])
                edf = edf.set_index(["symbol", "date"])
                # edf = edf.pivot(columns='symbol')
                # edf.swaplevel(i=0,j=1, axis=0)
                # edf.drop(columns=['symbol'])
                earnings_all_df = pd.concat([earnings_all_df, edf])
                # n_earnings = len(earnings)
                # logger.info(f"Total earnings reports for {ticker}: {n_earnings}")
        #    earliest_earn = earnings[-1] if len(earnings > 0 else 'None')
        #    logger.info(f"Earliest earnings report for {ticker}: {earliest_earn}")

        logger.info(f"Earnings sample report for {ticker}: \n{earnings}")
        earnings_all_df.index.names = ["Symbol", "Date"]
        earnings_all_df = earnings_all_df.sort_index()
        logger.info(
            f"Total number of earnings records for all stocks: \n{len(earnings_all_df)}"
        )
        logger.info(f"earnings_all_df: \n{earnings_all_df}")
        logger.info(
            f"len(earnings_all_df.index.levels[0]): \n{len(earnings_all_df.index.levels[0])}"
        )
        earnings_file = "data/data-3rd-party/earnings_calendar.parquet"
        earnings_all_df.to_parquet(earnings_file)
        ### Read back data and verify it
        tmp_earn_df = pd.read_parquet(earnings_file)
        logger.info(f"tmp_earn_df: \n{tmp_earn_df}")
        assert tmp_earn_df.index.names == ["Symbol", "Date"]
        assert len(tmp_earn_df) == len(earnings_all_df)
        assert sorted(tmp_earn_df.columns) == sorted(earnings_all_df.columns)
        logger.info(
            f"Sanity check passed for earnings data. Loaded OK from file: {earnings_file}"
        )

    def gather_stock_key_metrics(self):
        logger.info("Gathering key metrics data with company fundamentals...")
        keymetrics_all_df = pd.DataFrame()
        for ticker in self.stocks_ticker_set:
            kms = fmpsdk.key_metrics(
                apikey=self.FMP_API_KEY, symbol=ticker, period="quarter", limit=-1
            )
            if kms is not None and len(kms) > 0:
                kms_df = pd.DataFrame(kms)
                kms_df["date"] = pd.to_datetime(kms_df["date"])
                kms_df = kms_df.set_index(["symbol", "date"])
                # logger.info(f"Key metrics for {ticker} sample: \n{kms_df.columns}")
                keymetrics_all_df = pd.concat([keymetrics_all_df, kms_df])
                # logger.info(f"Key metrics concatenated {ticker}: \n{keymetrics_all_df.columns}")
                n_kms = len(kms_df)
                logger.info(f"Total key metrics reports for {ticker}: {n_kms}")
            else:
                logger.info(f"No {ticker} key metrics reports: kms={kms}")

        logger.info(f"keymetrics_all_df: \n{keymetrics_all_df}")
        logger.info(f"keymetrics_all_df.dtypes: \n{keymetrics_all_df.dtypes}")
        logger.info(f"len(keymetrics_all_df): \n{len(keymetrics_all_df)}")
        # index, row = next(keymetrics_all_df.iterrows())
        # row["averagePayables"]
        # prevent parquet serialization issues
        keymetrics_all_df["averagePayables"] = pd.to_numeric(
            keymetrics_all_df["averagePayables"], dtype_backend="pyarrow"
        )
        keymetrics_all_df.index.names = ["Symbol", "Date"]
        keymetrics_all_df = keymetrics_all_df.sort_index()
        kms_file = "data/data-3rd-party/keymetrics_history.parquet"
        keymetrics_all_df.to_parquet(kms_file, engine="pyarrow")
        ### Read back data and verify it
        temp_kms_df = pd.read_parquet(kms_file)
        logger.info(f"temp_kms_df: \n{temp_kms_df}")
        assert temp_kms_df.index.names == ["Symbol", "Date"]
        assert len(temp_kms_df) == len(keymetrics_all_df)
        assert sorted(temp_kms_df.columns) == sorted(keymetrics_all_df.columns)
        logger.info(
            f"Sanity check passed for key metrics data. Loaded OK from file: {kms_file}"
        )

    def gather_institutional_stock_ownership(self):
        logger.info("Gathering institutional ownership data...")
        inst_ownership_all_df = pd.DataFrame()
        for ticker in self.stocks_ticker_set:
            inst_ownership = institutional_symbol_ownership(
                apikey=self.FMP_API_KEY,
                symbol=ticker,
                limit=-1,
                includeCurrentQuarter=False,
            )
            # logger.info("inst_ownership: ", inst_ownership)
            if inst_ownership is not None and len(inst_ownership) > 0:
                inst_ownership_df = pd.DataFrame(inst_ownership)
                inst_ownership_df["date"] = pd.to_datetime(inst_ownership_df["date"])
                inst_ownership_df = inst_ownership_df.set_index(["symbol", "date"])
                # logger.info(f"Institutional ownership for {ticker} # columns: \n{len(inst_ownership_df.columns)}")
                n_iown = len(inst_ownership_df)
                logger.info(
                    f"Total institutional ownership reports for {ticker}: {n_iown}"
                )
                inst_ownership_all_df = pd.concat(
                    [inst_ownership_all_df, inst_ownership_df]
                )
                # logger.info(f"Institutional ownership concatenated {ticker} # columns: \n{inst_ownership_all_df.columns}")
            else:
                logger.info(
                    f"No {ticker} institutional ownership reports: inst_ownership={inst_ownership}"
                )
        inst_own_df = inst_ownership_all_df.copy()
        # prevent parquet serialization issues
        inst_own_df["totalInvestedChange"] = pd.to_numeric(
            inst_own_df["totalInvestedChange"],
            dtype_backend="pyarrow",
            downcast="integer",
        )

        def find_bad_cell():
            for index, row in inst_own_df.iterrows():
                try:
                    x = row["totalPutsChange"]
                    assert isinstance(x, int)
                except Exception as e:
                    logger.info(
                        f"Unable to convert to numeric type value:({x}), type({type(x)}), index({index}, \nerror:{e}\nrow:{row})"
                    )
                    break

        find_bad_cell()
        # type(inst_own_df["totalInvestedChange"][0])
        # clean up bad data from the third party source feed
        inst_own_df["totalInvestedChange"] = inst_own_df["totalInvestedChange"].astype(
            "float64"
        )
        inst_own_df["totalInvestedChange"] = inst_own_df["totalInvestedChange"].astype(
            "int64"
        )
        inst_own_df["cik"] = inst_own_df["cik"].replace("", -1)
        inst_own_df["cik"] = inst_own_df["cik"].astype("int64")
        inst_own_df["totalPutsChange"] = inst_own_df["totalPutsChange"].astype(
            "float64"
        )
        inst_own_df["totalPutsChange"] = inst_own_df["totalPutsChange"].astype("int64")
        # inst_own_df["totalPutsChange"] = pd.to_numeric(inst_own_df["totalPutsChange"], dtype_backend="pyarrow", downcast = 'integer')
        inst_own_df["totalCallsChange"] = inst_own_df["totalCallsChange"].astype(
            "float64"
        )
        inst_own_df["totalCallsChange"] = inst_own_df["totalCallsChange"].astype(
            "int64"
        )
        # inst_own_df.dtypes
        # inst_own_df
        inst_own_df.index.names = ["Symbol", "Date"]
        inst_own_df = inst_own_df.sort_index()
        # inst_own_df.index.names
        inst_ownership_file = (
            "data/data-3rd-party/institutional_symbol_ownership.parquet"
        )
        inst_own_df.to_parquet(inst_ownership_file, engine="pyarrow")

        ### Read back data and verify it
        tmp_int_own_df = pd.read_parquet(inst_ownership_file)
        logger.info(f"tmp_int_own_df: \n{tmp_int_own_df}")
        assert tmp_int_own_df.index.names == ["Symbol", "Date"]
        assert len(tmp_int_own_df) == len(inst_own_df)
        assert sorted(tmp_int_own_df.columns) == sorted(inst_own_df.columns)
        logger.info(
            f"Sanity check passed for institutional ownership data. Loaded OK from file: {inst_ownership_file}"
        )

    def gather_analyst_estimates(self):

        def _fetch_estimates(period=None):
            assert period in ["quarter", "annual"]
            estimates_all_df = pd.DataFrame()
            for ticker in self.stocks_ticker_set:  # ['ALTR']:
                est = analyst_estimates(
                    apikey=self.FMP_API_KEY, symbol=ticker, period=period, limit=-1
                )
                # logger.info('est:', est)
                if est is not None and len(est) > 0:
                    est_df = pd.DataFrame(est)
                    est_df["date"] = pd.to_datetime(est_df["date"])
                    est_df = est_df.set_index(["symbol", "date"])
                    # logger.info(f"Analyst estimates for {ticker} sample: \n{est_df.columns}")
                    estimates_all_df = pd.concat([estimates_all_df, est_df])
                    # logger.info(f"Key metrics concatenated {ticker}: \n{estimates_all_df.columns}")
                    n_est = len(est_df)
                    logger.info(
                        f"{n_est} total {ticker} {period} analyst estimates reports"
                    )
                else:
                    logger.info(
                        f"No {ticker} {period} analyst estimates reports: est={est}"
                    )

            return estimates_all_df

        logger.info("Gathering analyst estimates...")
        est_file_name_template = "data/data-3rd-party/analyst_estimates_{p}.parquet"
        for p in ["annual", "quarter"]:
            est_file_name = est_file_name_template.format(p=p)
            estimates_all_df = _fetch_estimates(p)
            estimates_all_df.index.names = ["Symbol", "Date"]
            estimates_all_df = estimates_all_df.sort_index()
            # est_file_name= f'data/analyst_estimates_{p}.csv.bz2'
            # estimates_all_df.to_csv(est_file_name)
            estimates_all_df.to_parquet(est_file_name)
            logger.info(f"Saved analyst estimates to: {est_file_name}")
            logger.info(f"all {p} estimates count: {len(estimates_all_df)}")
            # logger.info(f"{p} estimates sample:\n{estimates_all_df}")
            tmp_est_df = pd.read_parquet(est_file_name)
            assert tmp_est_df.index.names == ["Symbol", "Date"]
            assert len(tmp_est_df) == len(estimates_all_df)
            # logger.info(f"tmp_est_df.columns: {tmp_est_df.columns}")
            # logger.info(f"estimates_all_df.columns: {estimates_all_df.columns}")
            assert sorted(tmp_est_df.columns) == sorted(estimates_all_df.columns)
            logger.info(
                f"Sanity check passed for analyst estimates data. Loaded OK from file: {est_file_name}"
            )


# main function
def main():
    hfhub = HFHub()
    hfhub.download_data()
    g = MarketDataGatherer()
    g.gather_broad_market_data()
    g.gather_sectors_data()
    g.gather_stock_tickers()
    g.gather_stock_price_data()
    g.gather_earnings_data()
    g.gather_stock_key_metrics()
    g.gather_institutional_stock_ownership()
    g.gather_analyst_estimates()
    hfhub.upload_data()


if __name__ == "__main__":
    main()



"""CANSWIM Playground. A gradio app intended to be deployed on HF Hub."""

from canswim.model import CanswimModel
import pandas as pd
import gradio as gr
from canswim.hfhub import HFHub
import matplotlib.pyplot as plt
import random
import pandas as pd
from loguru import logger
from typing import Union, Optional, Sequence
import matplotlib

# pd.options.plotting.backend = "plotly"
# pd.options.plotting.backend = "matplotlib"
# pd.options.plotting.backend = "hvplot"

repo_id = "ivelin/canswim"


class CanswimPlayground:
    def get_target(self, ticker):
        target = self.canswim_model.targets.target_series[ticker]
        logger.info(
            f"target {ticker} start, end: {target.start_time()}, {target.end_time()}"
        )
        return target

    def get_past_covariates(self, ticker):
        past_covariates = self.canswim_model.covariates.past_covariates[ticker]
        logger.info(
            f"past covariates start, end: {past_covariates.start_time()}, {past_covariates.end_time()}"
        )
        return past_covariates

    def get_future_covariates(self, ticker):
        future_covariates = self.canswim_model.covariates.future_covariates[ticker]
        logger.info(
            f"future covariates start, end: {future_covariates.start_time()}, {future_covariates.end_time()}"
        )
        return future_covariates

    def __init__(self):
        self.canswim_model = CanswimModel()
        self.hfhub = HFHub()
        # cache backtest and forecast predictions per ticker
        self.forecast_cache = {}
        self.backtest_cache = {}

    def download_model(self):
        """Load model from HF Hub"""
        # download model from hf hub
        self.canswim_model.download_model(repo_id=repo_id)
        logger.info(f"trainer params {self.canswim_model.torch_model.trainer_params}")
        self.canswim_model.torch_model.trainer_params["logger"] = False

    def download_data(self):
        """Prepare time series for model forecast"""
        self.hfhub.download_data(repo_id=repo_id)
        # load raw data from hf hub
        start_date = pd.Timestamp("2019-10-21")
        self.canswim_model.load_data(start_date=start_date)
        # prepare timeseries for forecast
        self.canswim_model.prepare_forecast_data(start_date=start_date)
        self.tickers = self.canswim_model.targets_ticker_list

    def plot_quantiles_df(
        self,
        df: pd.DataFrame = None,
        new_plot: bool = False,
        central_quantile: Union[float, str] = 0.5,
        low_quantile: Optional[float] = 0.05,
        high_quantile: Optional[float] = 0.95,
        default_formatting: bool = True,
        label: Optional[Union[str, Sequence[str]]] = "",
        max_nr_components: int = 10,
        ax: Optional[matplotlib.axes.Axes] = None,
        *args,
        **kwargs,
    ) -> matplotlib.axes.Axes:
        """Plot the series saved as parquet via dataframe.
        This method recreates the darts.TimeSeries.plot() method, which does not currently support restoring stochastic timeseries after saving.

        This is a wrapper method around :func:`xarray.DataArray.plot()`.

        Parameters
        ----------
        new_plot
            Whether to spawn a new axis to plot on. See also parameter `ax`.
        central_quantile
            The quantile (between 0 and 1) to plot as a "central" value, if the series is stochastic (i.e., if
            it has multiple samples). This will be applied on each component separately (i.e., to display quantiles
            of the components' marginal distributions). For instance, setting `central_quantile=0.5` will plot the
            median of each component. `central_quantile` can also be set to 'mean'.
        low_quantile
            The quantile to use for the lower bound of the plotted confidence interval. Similar to `central_quantile`,
            this is applied to each component separately (i.e., displaying marginal distributions). No confidence
            interval is shown if `confidence_low_quantile` is None (default 0.05).
        high_quantile
            The quantile to use for the upper bound of the plotted confidence interval. Similar to `central_quantile`,
            this is applied to each component separately (i.e., displaying marginal distributions). No confidence
            interval is shown if `high_quantile` is None (default 0.95).
        default_formatting
            Whether to use the darts default scheme.
        label
            Can either be a string or list of strings. If a string and the series only has a single component, it is
            used as the label for that component. If a string and the series has multiple components, it is used as
            a prefix for each component name. If a list of strings with length equal to the number of components in
            the series, the labels will be mapped to the components in order.
        max_nr_components
            The maximum number of components of a series to plot. -1 means all components will be plotted.
        ax
            Optionally, an axis to plot on. If `None`, and `new_plot=False`, will use the current axis. If
            `new_plot=True`, will create a new axis.
        alpha
             Optionally, set the line alpha for deterministic series, or the confidence interval alpha for
            probabilistic series.
        args
            some positional arguments for the `plot()` method
        kwargs
            some keyword arguments for the `plot()` method

        Returns
        -------
        matplotlib.axes.Axes
            Either the passed `ax` axis, a newly created one if `new_plot=True`, or the existing one.
        """

        if df is None or len(df) == 0:
            logger.info("Dataframe is empty. Nothing to plot.")
            return

        alpha_confidence_intvls = 0.25

        if central_quantile != "mean":
            assert isinstance(central_quantile, float) and 0.0 <= central_quantile <= 1.0, 'central_quantile must be either "mean", or a float between 0 and 1.'

        if high_quantile is not None and low_quantile is not None:
            assert 0.0 <= low_quantile <= 1.0 and 0.0 <= high_quantile <= 1.0, "confidence interval low and high quantiles must be between 0 and 1."

        if new_plot:
            fig, ax = plt.subplots()
        else:
            if ax is None:
                ax = plt.gca()

        if not any(lw in kwargs for lw in ["lw", "linewidth"]):
            kwargs["lw"] = 2

        if label is not None:
            assert isinstance(label, str), "The label argument should be a string"
            custom_labels = True
        else:
            custom_labels = False

        if central_quantile == "mean":
            central_series = df.mean(axis=1)
        else:
            central_series = df.quantile(q=central_quantile, axis=1)

            alpha = kwargs["alpha"] if "alpha" in kwargs else None
            if custom_labels:
                label_to_use = label
            else:
                if label == "":
                    label_to_use = "Value"
                else:
                    label_to_use = label
            kwargs["label"] = label_to_use

            if len(central_series) > 1:
                p = ax.plot(
                    df.index.values,
                    central_series.values,
                    "-",
                    *args,
                    **kwargs
                )
            # empty TimeSeries
            elif len(central_series) == 0:
                p = ax.plot(
                    [],
                    [],
                    *args,
                    **kwargs,
                )
                ax.set_xlabel(df.index.name)
            else:
                p = ax.plot(
                    [df.index.min()],
                    central_series.values[0],
                    "o",
                    *args,
                    **kwargs,
                )
            color_used = p[0].get_color() if default_formatting else None

            # Optionally show confidence intervals
            if (
                low_quantile is not None
                and high_quantile is not None
            ):
                low_series = df.quantile(q=low_quantile, axis=1)
                high_series = df.quantile(q=high_quantile, axis=1)
                if len(low_series) > 1:
                    ax.fill_between(
                        df.index,
                        low_series,
                        high_series,
                        color=color_used,
                        alpha=(alpha if alpha is not None else alpha_confidence_intvls),
                    )
                else:
                    ax.plot(
                        [df.index.min(), df.index.min()],
                        [low_series.iloc[0], high_series.iloc[0]],
                        "-+",
                        color=color_used,
                        lw=2,
                    )

        ax.legend()
        return ax


    def plot_forecast(self, ticker: str = None, lowq: int = 0.2):
        target = self.get_target(ticker)
        saved_forecast_df_list = self.get_saved_forecast(ticker=ticker)
        lq = lowq / 100
        fig, axes = plt.subplots(figsize=(20, 12))
        target.plot(label=f"{ticker} Close actual")
        logger.info(f"Plotting saved forecast: {saved_forecast_df_list}")
        if saved_forecast_df_list is not None and len(saved_forecast_df_list) > 0:
            for forecast in saved_forecast_df_list:
                self.plot_quantiles_df(df=forecast, low_quantile=lq, high_quantile=0.95, label=f"{ticker} Close forecast")
        plt.legend()
        return fig

    def get_saved_forecast(self, ticker: str = None):
        """Load forecasts from storage to a list of individual forecast series with quantile sampling"""
        # load parquet partition for stock
        logger.info(f"Loading saved forecast for {ticker}")
        filters = [("symbol", "=", ticker)] # "AAON"
        df = pd.read_parquet(
            "data/forecast/",
            filters=filters,
        )
        logger.info(f"df columns count: {len(df.columns)}")
        logger.info(f"df row count: {len(df)}")
        logger.info(f"df columns: {df.columns}")
        logger.info(f"df column types: \n{df.dtypes}")
        logger.info(f"df row sample: {df}")
        df = df.drop(columns=["symbol"])
        df_list = []
        for y in df["forecast_start_year"].unique():
            for m in df["forecast_start_month"].unique():
                for d in df["forecast_start_day"].unique():
                    single_forecast = df.loc[(df["forecast_start_year"] == y) & (df["forecast_start_month"] == m) & (df["forecast_start_day"] == d)]
                    single_forecast = single_forecast.drop(columns=[
                        "forecast_start_year",
                        "forecast_start_month",
                        "forecast_start_day",])
                    single_forecast = single_forecast.sort_index()
                    df_list.append(single_forecast)
        return df_list


def main():
    with gr.Blocks() as demo:
        gr.Markdown(
            """
        CANSWIM Playground for CANSLIM style investors.
        * __NOT FINANCIAL OR INVESTMENT ADVICE. USE AT YOUR OWN RISK!__
        * Model trainer source repo [here](https://github.com/ivelin/canswim). Feedback welcome via github issues.
        """
        )

        canswim_playground = CanswimPlayground()
        canswim_playground.download_model()
        canswim_playground.download_data()

        with gr.Tab("Charts"):
            plotComponent = gr.Plot()
            with gr.Row():
                sorted_tickers = sorted(canswim_playground.tickers)
                logger.info("Dropdown tickers: ", sorted_tickers)
                tickerDropdown = gr.Dropdown(
                    choices=sorted_tickers,
                    label="Stock Symbol",
                    value=random.sample(sorted_tickers, 1)[0],
                )
                lowq = gr.Slider(
                    5,
                    80,
                    value=20,
                    label="Forecast probability low threshold",
                    info="Choose from 5% to 80%",
                )
            pass
        with gr.Tab("Scans"):
            pass
        with gr.Tab("Advanced Queries"):
            pass

        tickerDropdown.change(
            fn=canswim_playground.plot_forecast,
            inputs=[tickerDropdown, lowq],
            outputs=[plotComponent],
            queue=False,
        )
        lowq.change(
            fn=canswim_playground.plot_forecast,
            inputs=[tickerDropdown, lowq],
            outputs=[plotComponent],
            queue=False,
        )
        demo.load(
            fn=canswim_playground.plot_forecast,
            inputs=[tickerDropdown, lowq],
            outputs=[plotComponent],
            queue=False,
        )

    demo.launch()


if __name__ == "__main__":
    main()
